// Copyright(c)'1994-2009 by The Givaro group
// This file is part of Givaro.
// Givaro is governed by the CeCILL-B license under French law
// and abiding by the rules of distribution of free software.
// see the COPYRIGHT file for more details.

#ifndef __GIVARO_poly1_dense_ops_INL
#define __GIVARO_poly1_dense_ops_INL

#include "givaro/givpoly1cstor.inl"
#include "givaro/givpoly1misc.inl"
#include "givaro/givpoly1addsub.inl"
#include "givaro/givpoly1muldiv.inl"
#include "givaro/givpoly1kara.inl"
#include "givaro/givpoly1midmul.inl"
#include "givaro/givpoly1axpy.inl"
#include "givaro/givpoly1io.inl"
#include "givaro/givpoly1gcd.inl"
#include "givaro/givpoly1sqrfree.inl"
#include "givaro/givpoly1cyclo.inl"
#include "givaro/givpoly1ratrecon.inl"

#endif // __GIVARO_poly1_dense_ops_INL
/* -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
// vim:sts=4:sw=4:ts=4:et:sr:cino=>s,f0,{0,g0,(0,\:0,t0,+0,=s
